package eu.jonahbauer.chat.bot.database;

import eu.jonahbauer.chat.bot.database.annotations.Column;
import eu.jonahbauer.chat.bot.database.annotations.Table;
import org.intellij.lang.annotations.Language;
import org.jetbrains.annotations.NotNull;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.sql.SQLException;
import java.util.List;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;

public class DatabaseTest {

    @TempDir
    static Path temp;

    Database database;

    @BeforeEach
    void init() throws IOException, SQLException {
        var file = Files.createTempFile(temp, "database", ".db");
        var url = "jdbc:sqlite:" + file.toAbsolutePath();
        database = new Database(url);
        database.executeUpdate("""
        CREATE TABLE "pizza"
        (
            "id"            INTEGER PRIMARY KEY AUTOINCREMENT,
            "restaurant_id" INTEGER      NOT NULL,
            "number"        INTEGER      NOT NULL,
            "name"          VARCHAR(255) NOT NULL,
            "ingredients"   VARCHAR(255) NOT NULL,
            "price"         DOUBLE       NOT NULL,
            "test"          INTEGER
        )
        """);
    }

    @Test
    void crud() throws SQLException {
        crud(Pizza.class, Pizza::new);
    }

    @Test
    void annotations() throws SQLException {
        crud(AnnotatedPizza.class, AnnotatedPizza::new);
    }

    @Test
    void transaction() throws SQLException {
        assertThrowsExactly(RuntimeException.class, () -> {
            database.transaction(() -> {
                assertThrows(IllegalStateException.class, () -> database.transaction(() -> {}));

                database.insert(new Pizza(-1, 1, 10, "Salami", "Salami, Tomaten, Brot", 12.5));
                assertEquals(1, database.findAll(Pizza.class).size());
                throw new RuntimeException();
            });
        });
        assertEquals(List.of(), database.findAll(Pizza.class));
    }

    @Test
    void find() throws SQLException {
        var salami = new Pizza(1, 1, 10, "Salami", "Salami, Tomaten, Brot", 12.5);
        var speziale = new Pizza(2, 1, 15, "Speziale", "Salami, Schinken, Pilze, Tomaten, Brot", 12.5);

        assertEquals(1, database.insert(salami));
        assertEquals(2, database.insert(speziale));

        assertEquals(Optional.empty(), database.findWhere(Pizza.class, "name", "Hawaii"));
        assertEquals(Optional.of(salami), database.findWhere(Pizza.class, "name", "Salami"));
        assertEquals(Optional.of(speziale), database.findWhere(Pizza.class, "name", "Speziale"));
        assertThrows(NonUniqueResultException.class, () -> database.findWhere(Pizza.class, "restaurantId", 1L));
        assertEquals(List.of(salami, speziale), database.findAllWhere(Pizza.class, "restaurantId", 1L));
    }

    @Test
    void join() throws SQLException {
        var salami = new Pizza(1, 1, 10, "Salami", "Salami, Tomaten, Brot", 12.5);
        var speziale = new Pizza(2, 1, 15, "Speziale", "Salami, Schinken, Pilze, Tomaten, Brot", 12.5);

        assertEquals(1, database.insert(salami));
        assertEquals(2, database.insert(speziale));

        @Language("SQL")
        var query = """
                    SELECT "pizza".*
                    FROM "pizza"
                        JOIN "pizza" "p2" ON "pizza"."restaurant_id" = "p2"."restaurant_id"
                    """;
        assertEquals(List.of(salami, salami, speziale, speziale), database.executeQuery(Pizza.class, query));

        @Language("SQL")
        var query2 = """
                    SELECT "p2".*
                    FROM "pizza"
                        JOIN "pizza" "p2" ON "pizza"."restaurant_id" = "p2"."restaurant_id"
                    """;
        assertEquals(List.of(salami, speziale, salami, speziale), database.executeQuery(Pizza.class, query2));
    }

    private <T extends Record & Entity<T>> void crud(@NotNull Class<T> type, @NotNull PizzaBaker<T> baker) throws SQLException {
        var salami = baker.bake(-1, 1, 10, "Salami", "Salami, Tomaten, Brot", 12.5);
        var persistent = baker.bake(1, 1, 10, "Salami", "Salami, Tomaten, Brot", 12.5);
        var speziale = baker.bake(1, 1, 15, "Speziale", "Salami, Schinken, Pilze, Tomaten, Brot", 12.5);

        // create
        assertEquals(1, database.insert(salami));
        assertEquals(List.of(persistent), database.findAll(type));
        assertEquals(Optional.of(persistent), database.findById(type, 1));
        assertEquals(Optional.empty(), database.findById(type, 2));

        // update
        assertFalse(database.update(salami));
        assertTrue(database.update(speziale));
        assertEquals(List.of(speziale), database.findAll(type));
        assertEquals(Optional.of(speziale), database.findById(type, 1));
        assertEquals(Optional.empty(), database.findById(type, 2));

        // delete
        assertFalse(database.delete(type, 2));
        assertTrue(database.delete(type, 1));
        assertEquals(List.of(), database.findAll(type));

        // truncate
        database.insert(salami);
        database.insert(speziale);
        assertEquals(2, database.findAll(type).size());
        assertEquals(2, database.truncate(type));
        assertEquals(List.of(), database.findAll(type));
    }

    @AfterEach
    void cleanup() {
        database.close();
    }

    private interface PizzaBaker<T extends Record & Entity<T>> {
        @NotNull T bake(long id, long restaurantId, int number, @NotNull String name, @NotNull String ingredients, double price);
    }
    
    public record Pizza(long id, long restaurantId, int number, @NotNull String name, @NotNull String ingredients, double price) implements Entity<Pizza> {}
    
    @Table("pizza")
    public record AnnotatedPizza(long id, @Column("restaurant_id") long restaurant, int number, @NotNull String name, @NotNull String ingredients, double price) implements Entity<AnnotatedPizza> {}
}
