package eu.jonahbauer.chat.bot.pizza;

import eu.jonahbauer.chat.bot.api.Message.Post;
import eu.jonahbauer.chat.bot.pizza.model.*;
import eu.jonahbauer.chat.bot.pizza.model.table.*;
import eu.jonahbauer.chat.bot.test.MockChat;
import eu.jonahbauer.chat.bot.database.Database;
import eu.jonahbauer.chat.bot.test.ChatBotFactoryAccess;
import org.jetbrains.annotations.NotNull;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.sql.SQLException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Objects;

import static eu.jonahbauer.chat.bot.pizza.util.FormatUtils.*;
import static java.util.FormatProcessor.FMT;
import static org.junit.jupiter.api.Assertions.*;

class PizzaBotTest {
    @TempDir
    static Path temp;

    Database database;
    MockChat chat;
    PizzaBot bot;

    User admin;
    User user;
    User user2;

    Event event;
    Order order0; // alte bestellung (bestellannahmeschluss vorbei)
    Order order1; // alte bestellung
    Order order; // laufende bestellung

    Pizza salami;
    Pizza margherita;

    @BeforeEach
    void init() throws IOException, SQLException {
        var file = Files.createTempFile(temp, "database", ".db");
        database = new Database("jdbc:sqlite:" + file.toAbsolutePath());
        run("init.sql");
        run("restaurants/sonthofen_2023.sql");

        chat = new MockChat();
        bot = ChatBotFactoryAccess.create(() -> new PizzaBot(Collections.singleton("marek"), database));

        admin = database.save(new User(0, "Admin"));
        user = database.save(new User(1337, "User1"));
        user2 = database.save(new User(42, "User2"));

        event = database.save(new Event(0, "Veranstaltung", admin.id(), true));
        order0 = database.save(new Order(0, event.id(), 1, "28.03.2024", Instant.ofEpochMilli(System.currentTimeMillis() - 3_600_000), 90.0));
        order1 = database.save(new Order(0, event.id(), 1, "29.03.2024", Instant.ofEpochMilli(System.currentTimeMillis() + 3_600_000), null));
        order = database.save(new Order(0, event.id(), 1, "30.03.2024", Instant.ofEpochMilli(System.currentTimeMillis() + 7_200_000), null));

        salami = database.findWhere(Pizza.class, Pizza.Fields.name, "Salami").orElseThrow();
        margherita = database.findWhere(Pizza.class, Pizza.Fields.name, "Margherita").orElseThrow();
    }

    @AfterEach
    void print() {
        for (var message : chat.getMessages()) {
            System.out.println("\n" + message.name() + "\n" + message.message());
        }
    }

    private void run(@NotNull String script) throws SQLException, IOException {
        try (var conn = database.getConnection()) {
            var runner = new ScriptRunner(conn, false, true);
            runner.setLogWriter(null);
            try (var stream = Objects.requireNonNull(PizzaBot.class.getResourceAsStream(script))) {
                try (var reader = new InputStreamReader(stream, StandardCharsets.UTF_8)) {
                    runner.runScript(reader);
                }
            }
        }
    }

    private void deadline() throws SQLException {
        database.delete(order);
        database.delete(order1);
        order = order0;
    }

    private void post(@NotNull User user, @NotNull String message) {
        var post = new Post(0, user.username(), message, "marek", LocalDateTime.now(), 0L, user.id(), user.username(), "ffffff", 0);
        bot.onMessage(chat, post);
    }

    @Nested
    class OrderPizza {

        @Test
        void orderByPizzaName() throws SQLException {
            post(user, "!marek order SALAMI");

            var items = database.findAll(OrderItem.class);
            assertEquals(1, items.size());

            var item = items.getFirst();
            assertNull(item.notes());
            assertEquals(order.id(), item.orderId());
            assertEquals(salami.id(), item.pizzaId());

            var mappings = database.findAll(OrderItemToUser.class);
            assertEquals(List.of(new OrderItemToUser(item.id(), user.id(), false)), mappings);
        }

        @Test
        void orderByPizzaNameWithNotes() throws SQLException {
            post(user, "!marek order SALAMI Hello World");

            var items = database.findAll(OrderItem.class);
            assertEquals(1, items.size());

            var item = items.getFirst();
            assertEquals("Hello World", item.notes());
            assertEquals(order.id(), item.orderId());
            assertEquals(salami.id(), item.pizzaId());
        }

        @Test
        void orderByPizzaNameFailsAfterDeadline() throws SQLException {
            deadline();
            post(user, "!marek order SALAMI");

            var items = database.findAll(OrderItem.class);
            assertEquals(0, items.size());
        }

        @Test
        void orderByPizzaNumber() throws SQLException {
            post(user, "!marek order 13");

            var items = database.findAll(OrderItem.class);
            assertEquals(1, items.size());

            var item = items.getFirst();
            assertNull(item.notes());
            assertEquals(order.id(), item.orderId());

            var pizza = database.findById(Pizza.class, item.pizzaId());
            assertTrue(pizza.isPresent());
            assertEquals("Prosciutto", pizza.get().name());
        }

        @Test
        void orderAddById() throws SQLException {
            var item = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(admin, STR."!marek order add \{item.id()} \{user2.username()}");

            assertEquals(
                    List.of(
                            new OrderItemToUser(item.id(), user.id(), false),
                            new OrderItemToUser(item.id(), user2.id(), false)
                    ),
                    database.findAll(OrderItemToUser.class)
            );
        }

        @Test
        void orderAddByName() throws SQLException {
            var item = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(admin, STR."!marek order add \{user.username()} \{user2.username()}");

            assertEquals(
                    List.of(
                            new OrderItemToUser(item.id(), user.id(), false),
                            new OrderItemToUser(item.id(), user2.id(), false)
                    ),
                    database.findAll(OrderItemToUser.class)
            );
        }

        @Test
        void orderAddByNameFailsWhenNotUnique() throws SQLException {
            var item1 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item1.id(), user.id(), false));
            var item2 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item2.id(), user.id(), false));

            post(admin, STR."!marek order add \{user.username()} \{user2.username()}");

            assertEquals(
                    List.of(
                            new OrderItemToUser(item1.id(), user.id(), false),
                            new OrderItemToUser(item2.id(), user.id(), false)
                    ),
                    database.findAll(OrderItemToUser.class)
            );
        }

        @Test
        void orderAddByIdFailsWhenAlreadyPayed() throws SQLException {
            var item = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));
            database.insert(new OrderItemToUser(item.id(), user2.id(), true));

            post(admin, STR."!marek order add \{item.id()} \{admin.username()}");

            assertEquals(
                    List.of(
                            new OrderItemToUser(item.id(), user.id(), false),
                            new OrderItemToUser(item.id(), user2.id(), true)
                    ),
                    database.findAll(OrderItemToUser.class)
            );
        }
    }

    @Nested
    class Revoke {

        @Test
        void orderRevoke() throws SQLException {
            var item1 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item1.id(), user.id(), false));
            var item2 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item2.id(), user2.id(), false));

            post(user, "!marek order revoke");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item2), items);
        }

        @Test
        void orderRevokeFailsAfterDeadline() throws SQLException {
            deadline();

            var item = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user, "!marek order revoke");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item), items);
        }

        @Test
        void orderRevokeFailsWithMultipleItems() throws SQLException {
            var item1 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item1.id(), user.id(), false));
            var item2 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item2.id(), user.id(), false));

            post(user, "!marek order revoke");

            var items = database.findAll(OrderItem.class);
            assertEquals(2, items.size());
        }

        @Test
        void orderRevokeFailsWhenItemBelongsToOtherOrder() throws SQLException {
            var item = database.save(new OrderItem(0, order1.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user, "!marek order revoke");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item), items);
        }

        @Test
        void orderRevokeByNameWithMultipleItems() throws SQLException {
            var item1 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item1.id(), user.id(), false));
            var item2 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item2.id(), user.id(), false));

            post(user, "!marek order revoke SALAMI");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item1), items);
        }

        @Test
        void orderRevokeByNameFailsWhenItemBelongsToOtherUser() throws SQLException {
            var item = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user2, "!marek order revoke MARGHERITA");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item), items);
        }

        @Test
        void orderRevokeByNameFailsWhenItemBelongsToOtherOrder() throws SQLException {
            var item = database.save(new OrderItem(0, order1.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user, "!marek order revoke MARGHERITA");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item), items);
        }

        @Test
        void orderRevokeById() throws SQLException {
            var item = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user, "!marek order revoke " + item.id());

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(), items);
        }

        @Test
        void orderRevokeByIdFailsWhenItemBelongsToOtherUser() throws SQLException {
            var item = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user2, "!marek order revoke " + item.id());

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item), items);
        }

        @Test
        void orderRevokeByIdFailsWhenItemBelongsToOtherOrder() throws SQLException {
            var item = database.save(new OrderItem(0, order1.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item.id(), user.id(), false));

            post(user, "!marek order revoke " + item.id());

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item), items);
        }

        @Test
        void orderRevokeAll() throws SQLException {
            var item1 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            database.insert(new OrderItemToUser(item1.id(), user.id(), false));
            var item2 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item2.id(), user.id(), false));
            var item3 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            database.insert(new OrderItemToUser(item3.id(), user2.id(), false));

            post(user, "!marek order revoke --all");

            var items = database.findAll(OrderItem.class);
            assertEquals(List.of(item3), items);
        }
    }

    @Nested
    class Payment {
        OrderItem item;
        OrderItemToUser itemToUser;
        OrderItemToUser itemToUser2;

        @BeforeEach
        void init() throws SQLException {
            item = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            itemToUser = database.save(new OrderItemToUser(item.id(), user.id(), false));
            itemToUser2 = database.save(new OrderItemToUser(item.id(), user2.id(), false));
        }

        @Test
        void paymentConfirmById() throws SQLException {
            post(admin, "!marek payment confirm " + item.id());

            var mappings = database.findAll(OrderItemToUser.class);
            assertEquals(List.of(itemToUser.withPayed(true), itemToUser2.withPayed(true)), mappings);
        }

        @Test
        void paymentConfirmByIdFailsWhenNotAdmin() throws SQLException {
            post(user, "!marek payment confirm " + item.id());

            var mappings = database.findAll(OrderItemToUser.class);
            assertEquals(List.of(itemToUser, itemToUser2), mappings);
        }

        @Test
        void paymentConfirmByIdAndUser() throws SQLException {
            post(admin, "!marek payment confirm " + item.id() + " " + user.username());

            var mappings = database.findAll(OrderItemToUser.class);
            assertEquals(List.of(itemToUser.withPayed(true), itemToUser2), mappings);
        }

        @Test
        void paymentConfirmByIdAndUserFailsWhenNotAdmin() throws SQLException {
            post(user, "!marek payment confirm " + item.id() + " " + user.username());

            var mappings = database.findAll(OrderItemToUser.class);
            assertEquals(List.of(itemToUser, itemToUser2), mappings);
        }

        @Test
        void paymentConfirmByUser() throws SQLException {
            var item2 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            var item2ToUser = database.save(new OrderItemToUser(item2.id(), user.id(), false));
            var item2ToUser2 = database.save(new OrderItemToUser(item2.id(), user2.id(), false));

            post(admin, "!marek payment confirm --all " + user.username());

            var mappings = database.findAll(OrderItemToUser.class);
            assertEquals(
                    List.of(itemToUser.withPayed(true), itemToUser2, item2ToUser.withPayed(true), item2ToUser2),
                    mappings
            );
        }
    }

    @Nested
    class Pay {
        @Test
        void pay() throws SQLException {
            post(admin, "!marek pay 42.5");

            var order = database.findById(Order.class, PizzaBotTest.this.order.id());
            assertTrue(order.isPresent());
            assertEquals(42.5, order.get().payment());
        }

        @Test
        void payFailsWhenNotAdmin() throws SQLException {
            post(user, "!marek pay 42.5");

            var order = database.findById(Order.class, PizzaBotTest.this.order.id());
            assertTrue(order.isPresent());
            assertEquals(PizzaBotTest.this.order, order.get());
        }

        @Test
        void payWithOrderName() throws SQLException {
            post(admin, "!marek pay " + order0.name() + " 42.5");

            var order0 = database.findById(Order.class, PizzaBotTest.this.order0.id());
            assertTrue(order0.isPresent());
            assertEquals(42.5, order0.get().payment());

            var order = database.findById(Order.class, PizzaBotTest.this.order.id());
            assertTrue(order.isPresent());
            assertNull(order.get().payment());
        }
    }

    @Nested
    class CheckAndSummary {
        OrderItem item1;
        OrderItemToUser item1ToUser;

        OrderItem item2;
        OrderItemToUser item2ToUser;

        OrderItem item3;
        OrderItemToUser item3ToUser;

        OrderItem item4;
        OrderItemToUser item4ToUser2;

        OrderItem item5;
        OrderItemToUser item5ToUser;
        OrderItemToUser item5ToUser2;

        @BeforeEach
        void init() throws SQLException {
            item1 = database.save(new OrderItem(0, order0.id(), margherita.id(), null));
            item1ToUser = database.save(new OrderItemToUser(item1.id(), user.id(), false));

            item2 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            item2ToUser = database.save(new OrderItemToUser(item2.id(), user.id(), false));

            item3 = database.save(new OrderItem(0, order.id(), salami.id(), null));
            item3ToUser = database.save(new OrderItemToUser(item3.id(), user.id(), false));

            item4 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            item4ToUser2 = database.save(new OrderItemToUser(item4.id(), user2.id(), false));

            item5 = database.save(new OrderItem(0, order.id(), margherita.id(), null));
            item5ToUser = database.save(new OrderItemToUser(item5.id(), user.id(), false));
            item5ToUser2 = database.save(new OrderItemToUser(item5.id(), user2.id(), false));
        }

        @Test
        void check() {
            post(admin, "!marek check");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Rechnung für %s\{order.name()}
                       alle Angaben ohne Gewähr

                       %s\{user.username()}: %.2f\{1.5 * margherita.price() + salami.price()} €
                       %s\{user2.username()}: %.2f\{1.5 * margherita.price()} €
                       """.trim();
            assertEquals(expected, strip(message.message()));
        }

        @Test
        void checkTip() throws SQLException {
            database.update(order.withPayment(40d));

            post(admin, "!marek check");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Rechnung für %s\{order.name()} (mit Trinkgeld)
                       alle Angaben ohne Gewähr

                       %s\{user.username()}: %.2f\{1.5 * margherita.price() + salami.price()} € (25.19 €)
                       %s\{user2.username()}: %.2f\{1.5 * margherita.price()} € (14.81 €)
                       """.trim();
            assertEquals(expected, strip(message.message()));
        }

        @Test
        void checkPayed() throws SQLException {
            database.executeUpdate("UPDATE order_item_to_user SET payed = TRUE WHERE user_id = " + user.id());

            post(admin, "!marek check");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Rechnung für %s\{order.name()}
                       alle Angaben ohne Gewähr

                       %s\{user.username()}: %.2f\{1.5 * margherita.price() + salami.price()} €
                       %s\{user2.username()}: %.2f\{1.5 * margherita.price()} €
                       """.trim();
            assertEquals(expected, strip(message.message()));
            assertEquals(strikethrough("User1: " + tabular(1.5 * margherita.price() + salami.price(), 0) + " €"), message.message().split("\n")[3]);
        }

        @Test
        void checkPartialPayed() throws SQLException {
            database.executeUpdate("UPDATE order_item_to_user SET payed = TRUE WHERE order_item_id = " + item5.id());

            var user1Total = 1.5 * margherita.price() + salami.price();
            var user1Remaining = margherita.price() + salami.price();
            var user2Total = 1.5 * margherita.price();
            var user2Remaining = margherita.price();

            post(admin, "!marek check");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Rechnung für %s\{order.name()}
                       alle Angaben ohne Gewähr

                       %s\{user.username()}: %.2f\{user1Total} € %.2f\{user1Remaining} €
                       %s\{user2.username()}: %.2f\{user2Total} € %.2f\{user2Remaining} €
                       """.trim();
            assertEquals(expected, strip(message.message()));
            assertEquals(
                    STR."\{user.username()}: \{strikethrough(tabular(user1Total, 0)+ " €")} \{tabular(user1Remaining, 0)} €",
                    message.message().split("\n")[3]
            );
            assertEquals(
                    STR."\{user2.username()}: \{strikethrough(tabular(user2Total, 0)+ " €")} \{tabular(user2Remaining, 0)} €",
                    message.message().split("\n")[4]
            );
        }

        @Test
        void checkAllTip() throws SQLException {
            database.update(order.withPayment(40d));

            post(admin, "!marek check --all");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Rechnung für %s\{event.name()} (mit Trinkgeld)
                       alle Angaben ohne Gewähr

                       %s\{user.username()}: %.2f\{2.5 * margherita.price() + salami.price()} € (115.19 €)
                       %s\{user2.username()}: %.2f\{1.5 * margherita.price()} € (14.81 €)
                       """.trim();
            assertEquals(expected, strip(message.message()));
        }

        @Test
        void summary() {
            post(admin, "!marek summary");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Übersicht %s\{order.name()}

                       3x \{margherita.name().toUpperCase(Locale.ROOT)} (\{user.username()}, \{user.username()}+\{user2.username()}, \{user2.username()})
                       1x \{salami.name().toUpperCase(Locale.ROOT)} (\{user.username()})
                       """.trim();
            assertEquals(expected, strip(message.message()));
        }

        @Test
        void summaryAll() {
            post(admin, "!marek summary --all");
            var message = chat.getMessages().getFirst();
            var expected = FMT."""
                       Übersicht %s\{event.name()}

                       4x \{margherita.name().toUpperCase(Locale.ROOT)} (2x \{user.username()}, \{user.username()}+\{user2.username()}, \{user2.username()})
                       1x \{salami.name().toUpperCase(Locale.ROOT)} (\{user.username()})
                       """.trim();
            assertEquals(expected, strip(message.message()));
        }
    }

}
