package eu.jonahbauer.chat.server.management;

import eu.jonahbauer.chat.bot.config.BotConfig;
import lombok.*;
import org.jetbrains.annotations.NotNull;

import javax.management.openmbean.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Makes {@link BotConfig} accessible as JMX open data.
 */
@SuppressWarnings("unused")
@EqualsAndHashCode
public class BotConfigSupport implements CompositeDataView {
    private static final OpenType<String[]> STRING_ARRAY_TYPE;

    static {
        try {
            STRING_ARRAY_TYPE = ArrayType.getArrayType(SimpleType.STRING);
        } catch (OpenDataException e) {
            throw Lombok.sneakyThrow(e);
        }
    }

    private final @NotNull BotConfig config;

    public BotConfigSupport(@NotNull BotConfig config) {
        this.config = Objects.requireNonNull(config);
    }

    public @NotNull BotConfig unwrap() {
        return config;
    }


    // at least one getter is required for this class to be a valid open data type
    public String getType() {
        return config.getType();
    }

    // detected by java.management
    @SneakyThrows
    public static @NotNull BotConfigSupport from(@NotNull CompositeData composite) {
        var type = composite.getCompositeType();
        var keys = type.keySet();

        var out = BotConfig.builder();
        for (var key : keys) {
            var fieldType = type.getType(key);
            var value = composite.get(key);

            if (SimpleType.LONG.equals(fieldType)) {
                out.value(key, (Long) value);
            } else if (SimpleType.INTEGER.equals(fieldType)) {
                out.value(key, (Integer) value);
            } else if (SimpleType.SHORT.equals(fieldType)) {
                out.value(key, (Short) value);
            } else if (SimpleType.CHARACTER.equals(fieldType)) {
                out.value(key, (Character) value);
            } else if (SimpleType.BYTE.equals(fieldType)) {
                out.value(key, (Byte) value);
            } else if (SimpleType.DOUBLE.equals(fieldType)) {
                out.value(key, (Double) value);
            } else if (SimpleType.FLOAT.equals(fieldType)) {
                out.value(key, (Float) value);
            } else if (SimpleType.BOOLEAN.equals(fieldType)) {
                out.value(key, (Boolean) value);
            } else if (SimpleType.STRING.equals(fieldType)) {
                out.value(key, (String) value);
            } else if (STRING_ARRAY_TYPE.equals(fieldType)) {
                out.value(key, List.of((String[]) composite.get(key)));
            } else {
                throw new IllegalArgumentException("Unsupported field type: " + fieldType.getDescription());
            }
        }
        return new BotConfigSupport(out.build());
    }

    @Override
    public @NotNull CompositeData toCompositeData(@NotNull CompositeType ct) {
        try {
            var itemNames = new ArrayList<String>();
            var itemTypes = new ArrayList<OpenType<?>>();
            var values = new ArrayList<>();

            var keys = config.keySet();
            for (var key : keys) {
                var value = getOpenData(config.get(key).orElseThrow());
                var type = getOpenType(value);

                itemNames.add(key);
                itemTypes.add(type);
                values.add(value);
            }

            var xct = new CompositeType(
                    ct.getTypeName(),
                    ct.getDescription(),
                    itemNames.toArray(new String[0]),
                    itemNames.toArray(new String[0]),
                    itemTypes.toArray(new OpenType<?>[0])
            );

            return new CompositeDataSupport(xct, itemNames.toArray(new String[0]), values.toArray());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static @NotNull OpenType<?> getOpenType(@NotNull Object value) {
        return switch (value) {
            case Long _ -> SimpleType.LONG;
            case Double _ -> SimpleType.DOUBLE;
            case Boolean _ -> SimpleType.BOOLEAN;
            case String _ -> SimpleType.STRING;
            case String[] _ -> STRING_ARRAY_TYPE;
            default -> throw new IllegalArgumentException();
        };
    }

    @SuppressWarnings("SuspiciousToArrayCall")
    private static Object getOpenData(Object value) {
        return value instanceof List<?> list ? list.toArray(new String[0]) : value;
    }

    @Override
    public String toString() {
        return config.toString();
    }
}
