package eu.jonahbauer.chat.server.management.impl;

import eu.jonahbauer.chat.bot.impl.BotCreationException;
import eu.jonahbauer.chat.bot.impl.ChatBotFactory;
import eu.jonahbauer.chat.server.bot.ChatBotSupervisor;
import eu.jonahbauer.chat.server.management.BotConfigSupport;
import eu.jonahbauer.chat.server.management.ChatBotManagerMXBean;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Unmodifiable;

import javax.management.JMException;
import java.lang.management.ManagementFactory;
import java.util.*;
import java.util.stream.Collectors;

public final class ChatBotManager implements ChatBotManagerMXBean {
    private final @NotNull ChatBotSupervisor supervisor;

    public static void init(@NotNull ChatBotSupervisor supervisor) throws JMException {
        var impl = new ChatBotManager(supervisor);
        var server = ManagementFactory.getPlatformMBeanServer();
        server.registerMBean(new AdvancedMBean(impl), ChatBotManagerMXBean.NAME);
    }

    private ChatBotManager(@NotNull ChatBotSupervisor supervisor) {
        this.supervisor = Objects.requireNonNull(supervisor);
    }

    @Override
    public void start(@NotNull String name, @NotNull String type, @NotNull List<@NotNull String> channels) {
        try {
            supervisor.start(name, type, channels);
        } catch (BotCreationException ex) {
            throw new IllegalArgumentException(ex.getMessage(), ex.getCause());
        }
    }

    @Override
    public void start(@NotNull String name, @NotNull BotConfigSupport config) {
        try {
            supervisor.start(name, config.unwrap());
        } catch (BotCreationException ex) {
            throw new IllegalArgumentException(ex.getMessage(), ex.getCause());
        }
    }

    @Override
    public void stop(@NotNull String name) throws InterruptedException {
        supervisor.stop(name);
    }

    @Override
    public void stop() throws InterruptedException {
        supervisor.stop();
    }

    @Override
    public @Unmodifiable @NotNull Map<@NotNull String, @NotNull BotConfigSupport> getBots() {
        var out = new TreeMap<String, BotConfigSupport>();
        supervisor.getBots().forEach((key, value) -> out.put(key, new BotConfigSupport(value)));
        return Collections.unmodifiableMap(out);
    }

    @Override
    public @Unmodifiable @NotNull SortedSet<@NotNull String> getBotImplementations() {
        return Collections.unmodifiableSortedSet(
                ChatBotFactory.implementations().stream()
                        .map(Class::getCanonicalName)
                        .collect(Collectors.toCollection(TreeSet::new))
        );
    }
}
